"""Time-related Sphinx directives."""  # noqa: A005

from docutils import nodes
from sphinx.util.docutils import SphinxDirective
from sphinx_design.shared import create_component


class TimeDirective(SphinxDirective):
    """The TimeDirective directive is used to generate a time block."""

    has_content = False
    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = True

    def run(self):
        time = create_component(
            "ou-time", classes=["ou-time"], children=[nodes.Text(self.arguments[0], self.arguments[0])]
        )
        return [time]


def setup(app):
    """Setup the Time extensions."""
    app.add_directive("time", TimeDirective)



project = "OU Book Theme"
root_doc = "index"
extensions = ["myst_parser", "ou_book_theme", "sphinx_external_toc", "sphinxcontrib.mermaid"]
language = "en"
project_copyright = "2023-%Y"

numfig = True

html_title = "OU Book Theme"
html_theme = "ou_book_theme"

myst_enable_extensions = [
    "amsmath",
    "attrs_inline",
    "colon_fence",
    "deflist",
    "dollarmath",
    "fieldlist",
    "html_admonition",
    "html_image",
    "replacements",
    "smartquotes",
    "strikethrough",
    "substitution",
    "tasklist",
]



"""Additional Sphinx directives for the where-next blocks."""

from docutils import nodes
from sphinx.transforms.post_transforms import SphinxPostTransform
from sphinx.util.docutils import SphinxDirective
from sphinx_design.shared import create_component, is_component


class WhereNextDirective(SphinxDirective):
    """The WhereNextDirective directive is used to generate an "Where Next" block."""

    has_content = True
    required_arguments = 0
    optional_arguments = 0
    final_argument_whitespace = True

    def run(self):
        where_next = create_component("ou-where-next")
        self.state.nested_parse(self.content, self.content_offset, where_next)
        return [where_next]


class WhereNextHtmlTransform(SphinxPostTransform):
    """Transform where-next containers into the HTML specific AST structures."""

    default_priority = 199
    formats = ("html",)

    def run(self):
        """Run the transform"""
        document: nodes.document = self.document
        for node in document.findall(lambda node: is_component(node, "ou-where-next")):
            newnode = create_component(
                "ou-where-next",
                classes=["ou-where-next"],
            )
            newnode += nodes.raw("", '<p class="ou-where-next-title">Where next?</p>', format="html")
            newnode += node.children
            node.replace_self(newnode)


class WhereNext(nodes.Element):
    def __init__(self, rawsource="", *children, **attributes):
        super().__init__(rawsource=rawsource, *children, **attributes)  # noqa: B026
        self.set_class("ou-where-next")


class WhereNextTitle(nodes.Element):
    def __init__(self, rawsource="", *children, **attributes):
        super().__init__(rawsource=rawsource, *children, **attributes)  # noqa: B026
        self.set_class("ou-where-next-title")


class WhereNextContent(nodes.Element):
    def __init__(self, rawsource="", *children, **attributes):
        super().__init__(rawsource=rawsource, *children, **attributes)  # noqa: B026
        self.set_class("ou-where-next-content")


def setup(app):
    """Setup the Where Next extensions."""
    app.add_directive("where-next", WhereNextDirective)
    app.add_post_transform(WhereNextHtmlTransform)



"""Additional Sphinx directives for OU errata lists."""

import re

from docutils import nodes
from sphinx.util.docutils import SphinxDirective
from sphinx_design.shared import create_component, is_component


class ErratalistDirective(SphinxDirective):
    """The Erratalist directive is used to generate the list of all errate in the site."""

    def run(self):
        """Generate the basic structure."""
        return [create_component("ou-errata-list", rawtext=self.content)]


class ErratumDirective(SphinxDirective):
    """The ErratumDirective directive is used to generate an individual erratum."""

    has_content = True
    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = True

    def run(self):
        """Generate the erratum reference target and content."""
        erratum_id = f"erratum-{self.env.new_serialno('erratum')}"
        erratum_target = nodes.target("", "", ids=[erratum_id])

        erratum = create_component(
            "ou-erratum",
            classes=["ou-erratum", "margin"],
            rawtext=self.content,
        )
        content = nodes.admonition("", classes=["erratum"])
        content.children.append(
            nodes.paragraph(self.arguments[0], text=self.arguments[0], classes=["admonition-title"])
        )
        erratum.children.append(content)
        self.state.nested_parse(self.content, self.content_offset, content)

        if not hasattr(self.env, "errata_list"):
            self.env.errata_list = []
        self.env.errata_list.append(
            {
                "docname": self.env.docname,
                "target": erratum_target,
                "title": self.arguments[0],
                "group": self.arguments[0],
                "erratum": content,
            }
        )

        return [erratum_target, erratum]


def purge_doc_errata(app, env, docname):  # noqa: ARG001
    """Purge the environment errata_list for the docname."""
    if hasattr(env, "errata_list"):
        env.errata_list = [erratum for erratum in env.errata_list if erratum["docname"] != docname]


def merge_errata(app, env, docnames, other):  # noqa: ARG001
    """Merge an updated errata_list."""
    if not hasattr(env, "errata_list"):
        env.errata_list = []
    if hasattr(other, "errata_list"):
        env.errata_list.extend(other.errata_list)


def errata_sort_key(erratum):
    """Sort errata titles by group."""
    if re.match(r"[0-9]{1,2}\.[0-9]{1,2}\.[0-9]{4}", erratum["group"]):
        parts = erratum["group"].split(".")
        parts.reverse()
        return [int(part) for part in parts]
    else:
        return [erratum["group"]]


def transform_errata_lists(app, doctree, fromdocname):
    """Transform all ou-errata-list components into nested bulleted lists."""
    for node in doctree.findall(lambda node: is_component(node, "ou-errata-list")):
        errata_list = nodes.bullet_list()

        if hasattr(app.env, "errata_list"):
            app.env.errata_list.sort(key=errata_sort_key)
            app.env.errata_list.reverse()
            last_group = None
            group = None
            for erratum in app.env.errata_list:
                if last_group != erratum["group"]:
                    last_group = erratum["group"]
                    group_item = nodes.list_item()
                    group_para = nodes.paragraph()
                    group_para.append(nodes.strong(last_group, last_group))
                    group_item.append(group_para)
                    group = nodes.bullet_list()
                    group_item.append(group)
                    errata_list.append(group_item)
                erratum_node = nodes.reference("", "")
                erratum_node["refdocname"] = erratum["docname"]
                erratum_node["refuri"] = (
                    f"{app.builder.get_relative_uri(fromdocname, erratum['docname'])}#{erratum['target']['refid']}"
                )
                erratum_node["internal"] = True
                erratum_node.extend(app.env.titles[erratum["docname"]].children)

                para = nodes.paragraph()
                para.append(erratum_node)
                list_item = nodes.list_item()
                list_item.append(para)
                list_item.extend(erratum["erratum"].children[1:])
                group.append(list_item)
        else:
            para = nodes.paragraph("No errata have been published.", "No errata have been published.")
            list_item = nodes.list_item()
            list_item.append(para)
            errata_list.append(list_item)

        node.replace_self(errata_list)


def setup(app):
    """Setup the Errata extensions."""
    app.add_directive("errata-list", ErratalistDirective)
    app.add_directive("erratum", ErratumDirective)
    app.connect("doctree-resolved", transform_errata_lists)
    app.connect("env-purge-doc", purge_doc_errata)
    app.connect("env-merge-info", merge_errata)



from ou_book_theme.extensions import activity, errata, time, video, where_next


def setup(app):
    """Setup all node extensions."""
    activity.setup(app)
    errata.setup(app)
    time.setup(app)
    video.setup(app)
    where_next.setup(app)



# SPDX-FileCopyrightText: 2023-present Mark Hall <mark.hall@work.room3b.eu>
#
# SPDX-License-Identifier: MIT
"""The main OU Book Theme entry point."""

from os import path
from pathlib import Path

from sphinx.application import Sphinx
from sphinx.util.fileutil import copy_asset, copy_asset_file
from sphinx_book_theme import hash_assets_for_files

from ou_book_theme import extensions


def get_html_theme_path():
    """Return list of HTML theme paths."""
    parent = Path(__file__).parent.resolve()
    theme_path = parent / "theme" / "ou_book_theme"
    return theme_path


def hash_js_css_assets(app, pagename, templatename, context, doctree):  # noqa: ARG001
    """Add hashed CSS and JS assets."""
    assets = ["styles/ou-book-theme.css", "scripts/ou-book-theme.js"]
    hash_assets_for_files(assets, get_html_theme_path() / "static", context, app)


def copy_custom_files(app, exc):
    """Copy custom logo and favicon files."""
    if app.builder.format == "html" and not exc:
        staticdir = path.join(app.builder.outdir, "_static")
        ouheadersdir = path.join(app.builder.outdir, "ouheaders")
        copy_asset_file(str(get_html_theme_path() / "static/images/favicon.svg"), staticdir)
        copy_asset(str(get_html_theme_path() / "ouheaders"), ouheadersdir)


def setup(app: Sphinx):
    """Set up the theme and its extensions."""
    # Register theme
    theme_dir = get_html_theme_path()
    app.add_html_theme("ou_book_theme", theme_dir)
    app.add_js_file("scripts/ou-book-theme.js")

    app.connect("html-page-context", hash_js_css_assets)
    app.connect("build-finished", copy_custom_files)

    extensions.setup(app)

    return {"parallel_read_safe": True, "parallel_write_safe": True}



"""Additional Sphinx directives for OU activities."""

from docutils import nodes
from sphinx.transforms.post_transforms import SphinxPostTransform
from sphinx.util.docutils import SphinxDirective
from sphinx_design.shared import create_component, is_component


class ActivityDirective(SphinxDirective):
    """The ActivityDirective directive is used to generate an activity block."""

    has_content = True
    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = True

    def run(self):
        activity = create_component("ou-activity", rawtext=self.content)
        activity += create_component(
            "ou-activity-title", rawtext=self.arguments[0], children=[nodes.Text(self.arguments[0], self.arguments[0])]
        )
        self.state.nested_parse(self.content, self.content_offset, activity)
        return [activity]


class ActivityHtmlTransform(SphinxPostTransform):
    """Transform activity containers into the HTML specific AST structures."""

    default_priority = 198
    formats = ("html",)

    def run(self):
        """Run the transform"""
        document: nodes.document = self.document
        for node in document.findall(lambda node: is_component(node, "ou-activity")):
            newnode = create_component(
                "ou-activity",
                classes=["ou-activity"],
            )
            title_node = create_component(
                "ou-activity-title", classes=["ou-activity-title"], children=node.children[0].children
            )
            newnode += title_node
            newnode += node.children[1:]
            node.replace_self(newnode)


class ActivityLatexTransform(SphinxPostTransform):
    """Transform activity containers into the LaTeX specific AST structures."""

    default_priority = 198
    formats = ("latex",)

    def run(self):
        """Run the transform"""
        document: nodes.document = self.document
        for node in document.findall(lambda node: is_component(node, "ou-activity")):
            newnode = create_component(
                "ou-activity",
                classes=["ou-activity"],
            )
            title_node = nodes.paragraph()
            title_node += nodes.raw("", "\\textbf{", format="latex")
            title_node += node.children[0].children
            title_node += nodes.raw("", "}", format="latex")
            newnode += title_node
            newnode += node.children[1:]
            node.replace_self(newnode)


class ActivityAnswerDirective(SphinxDirective):
    """The ActivityAnswerDirective directive is used to generate an activity block."""

    has_content = True
    required_arguments = 0
    optional_arguments = 0
    final_argument_whitespace = True

    def run(self):
        activity_answer = create_component(
            "ou-activity-answer",
            classes=["ou-activity-answer"],
            rawtext=self.content,
        )
        self.state.nested_parse(self.content, self.content_offset, activity_answer)
        return [activity_answer]


class ActivityAnswerHtmlTransform(SphinxPostTransform):
    """Transform activity containers into the HTML specific AST structures."""

    default_priority = 199
    formats = ("html",)

    def run(self):
        """Run the transform"""
        document: nodes.document = self.document
        for node in document.findall(lambda node: is_component(node, "ou-activity-answer")):
            newnode = create_component(
                "ou-activity-answer",
                classes=["ou-activity-answer"],
            )
            newnode += nodes.raw("", "<hr/>", format="html")
            newnode += nodes.raw(
                "",
                '<button class="sd-btn sd-btn-info ou-toggle ou-toggle-hidden"><span class="ou-toggle-show">Show answer</span><span class="ou-toggle-hide">Hide answer</span></button>',  # noqa: E501
                format="html",
            )
            content_container = create_component(
                "ou-activity-answer-content", classes=["ou-activity-answer-content"], children=node.children
            )
            newnode += content_container
            node.replace_self(newnode)


class ActivityAnswerLatexTransform(SphinxPostTransform):
    """Transform activity containers into the LaTeX specific AST structures."""

    default_priority = 199
    formats = ("latex",)

    def run(self):
        """Run the transform"""
        document: nodes.document = self.document
        for node in document.findall(lambda node: is_component(node, "ou-activity-answer")):
            newnode = create_component(
                "ou-activity-answer",
                classes=["ou-activity-answer"],
            )
            newnode += nodes.raw("", "\\textcolor[RGB]{220,220,220}{\\rule{\\linewidth}{0.2pt}}", format="latex")
            answer_title = nodes.paragraph()
            answer_title += nodes.raw("", "\\textbf{Activity Answer}", format="latex")
            newnode += answer_title
            newnode += node.children
            node.replace_self(newnode)


def setup(app):
    """Setup the Activity extensions."""
    app.add_directive("activity", ActivityDirective)
    app.add_directive("activity-answer", ActivityAnswerDirective)
    app.add_post_transform(ActivityHtmlTransform)
    app.add_post_transform(ActivityAnswerHtmlTransform)
    app.add_post_transform(ActivityLatexTransform)
    app.add_post_transform(ActivityAnswerLatexTransform)



"""JupyterBook CLI extension."""

import os
from typing import Annotated

from livereload import Server, shell
from typer import Argument, Option, Typer

cli = Typer()


@cli.command()
def build(path: Annotated[str, Argument(help="Path to the OU book")]) -> None:
    """Serve the OU Book locally."""
    full_build = shell(f"sphinx-build --builder html --fresh-env docs {os.path.join(path, '_build', 'html')}")
    full_build()


@cli.command()
def serve(
    path: Annotated[str, Argument(help="Path to the OU book")],
    host: Annotated[str, Option(help="The host to serve the book at")] = "127.0.0.1",
    port: Annotated[int, Option(help="The port to serve the book at")] = 8000,
) -> None:
    """Serve the OU Book locally."""
    partial_build = shell(f"sphinx-build --builder html {path} {os.path.join(path, '_build', 'html')}")
    full_build = shell(f"sphinx-build --builder html --fresh-env docs {os.path.join(path, '_build', 'html')}")
    full_build()

    server = Server()
    server.watch(f"{path}/**/*.md", partial_build)
    server.watch(f"{path}/**/*.yml", full_build)
    server.watch(f"{path}/**/*.py", full_build)
    server.watch(f"{path}/**/*.png", full_build)
    server.watch(f"{path}/**/*.jpg", full_build)
    server.watch(f"{path}/**/*.jpeg", full_build)
    server.watch(f"{path}/**/*.svg", full_build)
    server.serve(root=f"{path}/_build/html", port=port, host=host)



"""Development server."""

import os

from livereload import Server, shell

port = 8000
host = "0.0.0.0"  # noqa: S104

partial_build = shell(f"sphinx-build --builder html docs {os.path.join('docs', '_build', 'html')}")
full_build = shell(f"sphinx-build --builder html --fresh-env docs {os.path.join('docs', '_build', 'html')}")
build_theme = shell("npm run build")

build_theme()
full_build()

server = Server()
server.watch(os.path.join("docs", "**", "*.md"), partial_build)
server.watch(os.path.join("docs", "**", "*.yml"), full_build)
server.watch(os.path.join("docs", "**", "*.py"), full_build)
server.watch(os.path.join("docs", "**", "*.png"), full_build)
server.watch(os.path.join("docs", "**", "*.jpg"), full_build)
server.watch(os.path.join("docs", "**", "*.jpeg"), full_build)
server.watch(os.path.join("ou_book_theme", "assets", "**", "*.scss"), build_theme)
server.watch(os.path.join("ou_book_theme", "assets", "**", "*.js"), build_theme)
server.watch(os.path.join("ou_book_theme", "theme", "**", "*.*"), full_build)
server.serve(root=os.path.join("docs", "_build", "html"), port=port, host=host)



"""Additional Sphinx directives for OU OCL Videos."""

import urllib3
import webvtt
from docutils import nodes
from sphinx.transforms.post_transforms import SphinxPostTransform
from sphinx.util import logging
from sphinx.util.docutils import SphinxDirective
from sphinx_design.shared import create_component, is_component

logger = logging.getLogger(__name__)


class VideoDirective(SphinxDirective):
    """The VideoDirective directive is used to generate video block for videos."""

    has_content = False
    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = True

    def run(self):
        if self.arguments[0].startswith("https://video.ocl.open.ac.uk/ui/#/embed/"):
            video_id = self.arguments[0][40:]
            response = urllib3.request("GET", f"https://video.ocl.open.ac.uk/api/videos/{video_id}")
            if response.status == 200:  # noqa: PLR2004
                video = create_component("ou-video", rawtext=self.content, video_id=video_id)
                video += nodes.paragraph(
                    response.json()["title"],
                    response.json()["title"],
                    classes=["admonition-title"],
                )
                cdn_base_url = f"https://video-cdn.ocl.open.ac.uk/videos/{'/'.join(video_id)}/"
                response = urllib3.request("GET", f"{cdn_base_url}transcript.vtt")
                if response.status == 200:  # noqa: PLR2004
                    transcript = create_component(
                        "dropdown",
                        opened=False,
                        type="dropdown",
                        has_title=True,
                        icon="note",
                        chevron=True,
                        container_classes=["sd-mb-3", "ou-transcript"],
                        title_classes=[],
                        body_classes=[],
                    )
                    transcript += nodes.paragraph("Transcript", "Transcript")
                    buffer = []
                    for caption in webvtt.from_string(response.data.decode()):
                        buffer.append(caption.text.strip())
                        if caption.text.strip().endswith(".") and len(buffer) > 0:
                            transcript += nodes.paragraph(" ".join(buffer), " ".join(buffer))
                            buffer = []
                    if len(buffer) > 0:
                        video += nodes.paragraph(" ".join(buffer), " ".join(buffer))
                    video += transcript
                return [video]
            else:
                logger.error(f"OCL Video {video_id} not found")
        return []


class VideoHtmlTransform(SphinxPostTransform):
    """Transform video containers into the HTML specific AST structures."""

    default_priority = 198
    formats = ("html",)

    def run(self):
        """Run the transform"""
        document: nodes.document = self.document
        for node in document.findall(lambda node: is_component(node, "ou-video")):
            video_id = node["video_id"]
            newnode = create_component(
                "ou-video",
                classes=["admonition ou-video"],
            )
            newnode += node.children[0]
            newnode += nodes.raw(
                "",
                f'<iframe src="https://video.ocl.open.ac.uk/ui/#/embed/{video_id}" allow="fullscreen"></iframe>',
                format="html",
            )
            newnode += node.children[1:]
            node.replace_self(newnode)


class VideoLatexTransform(SphinxPostTransform):
    """Transform video containers into the LaTeX specific AST structures."""

    default_priority = 198
    formats = ("latex",)

    def run(self):
        """Run the transform"""
        document: nodes.document = self.document
        for node in document.findall(lambda node: is_component(node, "ou-video")):
            video_id = node["video_id"]
            newnode = create_component(
                "ou-video",
                classes=["admonition ou-video"],
            )
            video_title = nodes.paragraph()
            video_title += nodes.raw("", "\\textbf{Video: ", format="latex")
            video_title += node.children[0].children
            video_title += nodes.raw("", "}", format="latex")
            newnode.append(video_title)
            video_link_para = nodes.paragraph()
            video_link_para += nodes.Text("You can view this video at: ")
            video_link_para += nodes.reference(
                "",
                f"https://video.ocl.open.ac.uk/ui/#/embed/{video_id}",
                refuri=f"https://video.ocl.open.ac.uk/ui/#/embed/{video_id}",
            )
            newnode += video_link_para
            transcript_para = nodes.paragraph()
            transcript_para += nodes.raw("", "\\textbf{Transcript}", format="latex")
            newnode += transcript_para
            newnode += node.children[1].children[1:]
            newnode += nodes.raw("", "\\textcolor[RGB]{220,220,220}{\\rule{\\linewidth}{0.2pt}}", format="latex")
            node.replace_self(newnode)


def setup(app):
    """Setup the Activity extensions."""
    app.add_directive("video", VideoDirective)
    app.add_post_transform(VideoHtmlTransform)
    app.add_post_transform(VideoLatexTransform)
